#                Perforce Defect Tracking Integration Project
#                 <http://www.ravenbrook.com/project/p4dti/>
#
#                       TEST_P4DTI.PY -- TEST THE P4DTI
#
#                 Gareth Rees, Ravenbrook Limited, 2001-03-14
#
#
# 1. INTRODUCTION
#
# This test script tests the P4DTI.  Currently it only tests the TeamTrack
# integration on Windows NT, but it is designed to be extendable to other
# integrations and other operating systems.
#
# This script contains a lot of tricky code (for example, using the hostname or
# operating system name to pick a module to import or a class to use; or
# overriding a method in another module in order to snoop on its behaviour).
# Please take care when editing it.
#
# It uses the PyUnit unit test framework [PyUnit].
#
# The intended readership is project developers.
#
# This document is not confidential.
#
#
# 1.1. Regression tests in this script
#
# The section * means that the defect is tested throughout as a simple
# consequence of running the script; there is no particular test for it.
#
# Job    Section  Title
# -----------------------------------------------------------------------------
# job000003  6.3  Can't distinguish between job and Perforce error
# job000004  *    "p4 -G jobspec -o" doesn't work
# job000005  *    May fail on startup in future releases of TeamTrack
# job000037  6    Consistency checker script is inadequate
# job000047  8    Historical bugs are replicated but uninteresting
# job000048  6    Consistency checking of the configuration is inadequate
# job000050  8    There's no way to re-start replication
# job000053  10   Implicit changes made by the DT don't get replicated back
# job000057  *    Special chars in single-select keywords make replicator barf
# job000075  6    No automatic check of configuration
# job000107  9    You can't replicate an issue's project
# job000111  7    Replicator destroys jobspec that exists before it's installed
# job000112  9    Can't easily replicate by project
# job000116  6    Bugzilla integration doesn't do enough checking
# job000140  6.3  Logical field name "code" not allowed in TeamTrack
# job000158  6.2  Obscure error if Perforce can't be reached
# job000168  6.2  Easy to set dbms_port to a string
# job000170  6.2  Can't send e-mail if default replicator_address is unchanged
# job000181  *    Assertion failure in translate_1_to_0
# job000182  10   Elapsed time fields aren't replicated properly
# job000190  3.1  Connection to TeamTrack hangs for several minutes
# job000199  4.5  Auxiliary scripts send e-mail
# job000202  6.2  Errors from Perforce not reported well
# job000219  7    Jobs may become unusable when the jobspec is changed
# job000221  8    Refreshing Perforce jobs fails in Bugzilla integration
# job000233  10   If you submit a new issue to TeamTrack it overwrites the job
# job000240  7    Perforce can crash if P4DTI is started with existing jobs

import sys
sys.path.append('../code/replicator')
import cgi
import copy
import httplib
import logger
import message
import os
import p4
import re
import socket
import string
import tempfile
import time
import types
import unittest
import urllib


# 2. CONFIGURATION
#
#
# 2.1. Limitations
#
# This script can only support one basic configuration on each host (see
# section 2.2).  Tests of other configurations have to be made by changing
# configuration parameters in the script.
#
# This script can only test one defect tracker on each host (you have to change
# dt_name in config_HOSTNAME.py to test another defect tracker).
#
# Could it use an environment variable to work out which configuration to use?
#
#
# 2.2. Using a test configuration
#
# I want to be able to test different configurations, changes to
# configurations, and to check whether erroneous configurations are spotted
# correctly.
#
# I expect to find a working configuration in config_HOSTNAME.py.
# This configuration should be copied when a test is started, changes
# made to the copy, and the copy installed in sys.modules['config'] so
# that the P4DTI will use the test configuration instead of the one in
# config.py.  Note that HOSTNAME is the first component of the FQDN
# (e.g. 'swan' for 'swan.ravenbrook.com').
#
# I also expect to find a Perforce license in the location given by
# the p4_license_file variable in the config_HOSTNAME.py file.

config_module = __import__('config_' +
                           string.split(socket.gethostname(),'.')[0])
original_configuration = copy.copy(config_module.__dict__)
sys.modules['config'] = config_module
import config

def reset_configuration():
    # Delete current config, just in case we added something.
    for k in sys.modules['config'].__dict__.keys():
        if not re.match('^__.*__$', k):
            del sys.modules['config'].__dict__[k]
    # Restore old config.
    for (k,v) in original_configuration.items():
        sys.modules['config'].__dict__[k] = v


# 3. DEFECT TRACKER AND OPERATING INTERFACES
#
# Many of the test cases are generic: they don't depend on a particular defect
# tracker or operating system.  But they need an interface to the defect
# tracker in order to restart it, and to the operating system in order to start
# a new Perforce server.
#
# Each interface class should be called DT_OS_interface (where DT is
# config.dt_name and OS is os.name) is and must define these methods:
#
# restart_defect_tracker()
# restart_perforce()
#
# Note that there are no corresponding "stop" methods.  The tests leave
# Perforce and the defect tracker running so that failures can be investigated
# without all the evidence having disappeared.


# 3.1. Interface to TeamTrack on Windows NT

class TeamTrack_nt_interface:

    # The TeamTrack server.
    server = None

    # Temporary directory for files created by this test.
    tempdir = None


    # 3.1.1. Clean up TeamTrack
    #
    # I would prefer to create a new TeamTrack database each time this test is
    # run (so that I can check that extra fields are added correctly to the
    # CASES table, for example).  However, I don't know of a way to do that;
    # TeamTrack is controlled from the TeamTrack Administrator and I have no
    # idea how to automate it.
    #
    # So the next best thing to do is to clean out all P4DTI data from a
    # running TeamTrack server.  This means setting all P4DTI_ fields in the
    # CASES table to the empty string, and deleting all records from the
    # VCACTIONS table.
    #
    # This isn't ideal: each new test ends up working with the set of cases
    # that resulted from the previous test.

    def restart_defect_tracker(self):
        # Connect to the TeamTrack server.  This should take no more than 5
        # seconds (regression test for job000190).
        import teamtrack
        connection_start = time.time()
        self.server = teamtrack.connect(config.teamtrack_user,
                                        config.teamtrack_password,
                                        config.teamtrack_server)
        connection_time = time.time() - connection_start
        assert connection_time <= 10, \
               ("Connection to TeamTrack took %d seconds" % connection_time)

        # Empty the P4DTI_* fields in the CASES table, if there are any.
        cases = self.server.query(teamtrack.table['CASES'],'')
        fields = ['P4DTI_JOBNAME', 'P4DTI_RID', 'P4DTI_SID']
        for c in cases:
            changed = 0
            for f in fields:
                if c.has_key(f) and c[f] != '':
                    c[f] = ''
                    changed = 1
            if changed:
                c.update()

        # Delete all records in the VCACTIONS table.
        vcactions = self.server.query(teamtrack.table['VCACTIONS'],'')
        for v in vcactions:
            self.server.delete_record(teamtrack.table['VCACTIONS'], v['ID'])


    # 3.1.2. Kill old Perforce servers
    #
    # Find and terminate all running Perforce servers, if any.  This code is
    # based on the script "killProcName.py" that comes with the Python Win32
    # extensions.

    def kill_perforce(self):
        import win32pdhutil
        import win32api
        import win32con
	try:
            win32pdhutil.GetPerformanceAttributes('Process','ID Process',"p4d")
	except:
            pass
        pids = win32pdhutil.FindPerformanceAttributesByName("p4d")
        for pid in pids:
            h = win32api.OpenProcess(win32con.PROCESS_TERMINATE, 0, pid)
            win32api.TerminateProcess(h,0)
            win32api.CloseHandle(h)


    # 3.1.3. Start a new Perforce server
    #
    # Make a completely fresh Perforce server, with a new repository.

    def start_perforce(self):
        # Make a new repository directory.
        self.tempdir = tempfile.mktemp()
        os.mkdir(self.tempdir)

        # Copy the license.
        src = open(config.p4_license_file, 'r')
        dest = open(self.tempdir + '\\license', 'w')
        dest.writelines(src.readlines())
        src.close()
        dest.close()

        # Work out Perforce's port number and start a Perforce server.
        match = re.match(".*:([0-9]+)$", config.p4_port)
        if match:
            port = int(match.group(1))
        else:
            port = 1666
        import win32api
        win32api.WinExec("%s -p 0.0.0.0:%d -r %s"
                         % (config.p4_server_executable, port, self.tempdir))


    # 3.1.4. Restart Perforce
    #
    # By killing the old server and starting a new one.

    def restart_perforce(self):
        self.kill_perforce()
        self.start_perforce()



# 3.2. Interface to TeamTrack on Windows NT

class Bugzilla_posix_interface:

    # The Bugzilla server.
    server = None

    # Temporary directory for files created by this test.
    tempdir = None


    # 3.2.1. Restore Bugzilla
    #
    # Wipe out the existing Bugzilla database, if any, and restore a
    # known working one from a MySQL dump file.

    def restart_defect_tracker(self):
        os.system("mysqladmin -u bugs --force drop bugs > /dev/null")
        os.system("mysqladmin -u bugs create bugs > /dev/null")
        os.system("mysql -u bugs bugs < bugzilla.mysqldump > /dev/null")


    # 3.2.2. Kill running Perforce servers
    #
    # If there are any Perforce servers running on the magic port,
    # use p4 admin to kill them.

    def kill_running_perforce_servers(self):
        os.system("p4 -p %s -u %s admin stop > /dev/null"
                  % (config.p4_port, config.p4_user))

    # 3.2.3. Start a new Perforce server
    #
    # Make a completely fresh Perforce server, with a new repository.

    def start_perforce(self):
        # Make a new repository directory.
        self.tempdir = tempfile.mktemp()
        os.mkdir(self.tempdir)

        # Copy the license.
        src = open(config.p4_license_file, 'r')
        dest = open(self.tempdir + '/license', 'w')
        dest.writelines(src.readlines())
        src.close()
        dest.close()

        # Work out Perforce's port number and start a Perforce server.
        match = re.match(".*:([0-9]+)$", config.p4_port)
        if match:
            port = int(match.group(1))
        else:
            port = 1666
        os.system("%s -d -p 127.0.0.1:%d -r %s > /dev/null"
                  % (config.p4_server_executable, port, self.tempdir))


    # 3.2.4. Restart Perforce
    #
    # By killing the old server and starting a new one.

    def restart_perforce(self):
        self.kill_running_perforce_servers()
        self.start_perforce()



# 4. P4DTI TEST CASE BASE CLASS
#
# The p4dti_base class is a generic P4DTI test case.  It defines methods for
# setting up the integration and some utilities for recording and interrogating
# the output of the replicator.
#
# Other P4DTI test cases will inherit from p4dti_base.
#
# When a class implements several test cases, the methods that implement test
# cases (in the PyUnit sense) should have names starting "test_".  When a class
# implements a single test case, the method should be called "runTest".

class p4dti_base(unittest.TestCase):

    # Defect tracker interface (an instance of one of the classes in section 3
    # above).
    dti = eval(config.dt_name + '_' + os.name + '_interface')()

    # Messages written to the replicator's log.
    log_messages = []

    # Mail messages sent by the replicator.  Each message is a triple
    # (recipients, subject, body); see the definition of mail() method in
    # the replicator class.
    mail_messages = []

    # The replicator.
    r = None


    # 4.1. Snoop on the logger
    #
    # I want to be able to test that the correct messages are appearing in the
    # log, so I override the log method in the logger.multi_logger class so
    # that it just records the messages without printing them.  Debugging
    # messages are not recorded.
    #
    # The clear_log method can be used to clear this record of the log before
    # carrying out a test.

    def snoop_logger(self):
        logger.multi_logger.log = self.log
        logger.file_logger.log = self.log
        logger.sys_logger.log = self.log

    def log(self, msg):
        self.log_messages.append(msg)

    def clear_log(self):
        self.log_messages = []


    # 4.2. Snoop on e-mail

    def snoop_mail(self):
        import replicator
        replicator.mail = self.mail

    def mail(self, to, subject, body):
        self.mail_messages.append((to, subject, body))

    def clear_mail(self):
        self.mail_messages = []


    # 4.3. Check that the log is as expected
    #
    # These methods make various checks on the messages recorded in the log.

    def log_message_ids(self):
        return map(lambda m: m.id, self.log_messages)

    def expected_only(self, expected):
        for m in self.log_messages:
            assert m.id in expected, \
                   ("Unexpected message %d (%s) found in log %s"
                    % (m.id, str(m), self.log_message_ids()))

    def expected_only_in_range(self, min, max, expected):
        for m in self.log_messages:
            if m.id >= min and m.id <= max:
                assert m.id in expected, \
                       ("Unexpected message %d (%s) found in log %s"
                        % (m.id, str(m), self.log_message_ids()))

    def expected_not(self, unexpected):
        for m in self.log_messages:
            assert m.id not in unexpected, \
                   ("Unexpected message %d (%s) found in log %s"
                    % (m.id, str(m), self.log_message_ids()))

    def expected(self, expected):
        found = {}
        for m in self.log_messages:
            if found.has_key(m.id):
                found[m.id] = found[m.id] + 1
            else:
                found[m.id] = 1
        for id in expected:
            assert found.has_key(id) and found[id] > 0, \
                   ("Expected message %d not found in log %s"
                    % (id, self.log_message_ids()))
            found[id] = found[id] - 1


    # 4.4. Set up everything so that test cases can run
    #
    # Get a fresh configuration, a new defect tracker and a new Perforce
    # server.

    def setup_everything(self):
        reset_configuration()
        self.dti.restart_defect_tracker()
        self.dti.restart_perforce()
        self.snoop_logger()
        self.snoop_mail()

        # Get a Perforce interface.
        self.p4 = p4.p4( port = config.p4_port,
                         user = config.p4_user,
                         password = config.p4_password,
                         client_executable = config.p4_client_executable,
                         logger = self )


    # 4.5. Initialize the replicator
    #
    # Load the init module.  If it's already loaded, reload it: we must reload
    # it because we want the init module to run again and to pick up on the new
    # configuration.

    def initialize_replicator(self):
        self.clear_mail()
        if sys.modules.has_key('init'):
            reload(sys.modules['init'])
        import init
        self.r = init.r
        # Regression test for job000199.
        assert self.mail_messages == []


    # 4.6. Normal replicator startup
    #
    # This is a pseudo test case.  It checks that the replicator can start up
    # normally.  Other tests may depend on this having been run; for example,
    # the check_nothing test (5.1).  This test must be run before any other
    # test cases, and therefore can't be part of a test suite (the tests of
    # which may be run singly or in any order).  It should therefore be run
    # from the setUp() method of a subclass of this class (unless of course
    # that subclass has a different idea of what should happen at startup).

    def check_startup(self):
        self.initialize_replicator()

        # Expect to set up issues and replicate them, but no jobs or conflicts.
        self.clear_log()
        self.r.poll()
        self.expected_only_in_range(800, 1000, [845, 803, 804, 812])
        self.expected_not([846, 847, 806, 811])

        # Expect to see no issues replicated, and no issues actually changed.
        self.clear_log()
        self.r.poll()
        self.expected_not([844, 845, 803, 804, 812, 824])


    # 4.7. Exceptional replicator startup
    #
    # This method starts the replicator but expects to get the exception given
    # by 'error', with message id 'msgid'.

    def check_startup_error(self, error, msgid):
        try:
            self.initialize_replicator()
        except:
            (err, msg, _) = sys.exc_info()
            assert err == error, \
                   ("Expected error %s but got %s." % (error, err))
            assert msg.id == msgid, \
                   ("Expected message %d but got %d (%s)"
                    % (msgid, msg.id, str(msg)))
        else:
            self.fail("Expected error %s but didn't get it." % error)


    # 4.8. Check consistency
    #
    # This method checks that the databases are consistent.

    def check_consistency(self):
        self.clear_log()
        self.r.check_consistency()
        self.expected_only_in_range(800, 1000, [871, 883, 884, 885])


    # 4.9. Check replication of a single issue

    def check_replication_dt_to_p4(self, first_time = 0):
        self.clear_log()
        self.r.poll()
        self.expected([844, 804, 812])
        if first_time:
            self.expected([803]) # set up for replication
        self.expected_only_in_range(800, 1000, [844, 803, 804, 812])

        # Nothing should come back from Perforce.
        self.clear_log()
        self.r.poll()
        self.expected_not([844, 845, 803, 804, 812, 824])
        self.expected_only_in_range(800, 1000, [846, 805, 825])


# 4.10 Variant class
#
# This class has variant methods for each defect tracker (e.g., TeamTrack,
# Bugzilla) and calls the appropriate one.

class p4dti_variant(p4dti_base):
    def runTest(self):
	try:
	    test = getattr(self, config.dt_name)
	except AttributeError:
            assert 0, "No test variant for " + config.dt_name + "."
        test()


# 5. TEST CASES: NORMAL OPERATION

class normal(p4dti_base):
    def setUp(self):
        self.setup_everything()
        self.check_startup()


    # 5.1. If nothing has changed, then nothing happens

    def test_nothing(self):
        self.clear_log()
        self.r.poll()
        self.expected_only_in_range(800, 1000, [])


    # 5.2. The databases are consistent

    def test_consistency(self):
        self.check_consistency()


# 6. TEST CASES: INCORRECT CONFIGURATIONS
#
# This is a regression test of job000037, job000075 and job000116.

class bogus_config(p4dti_base):
    def setUp(self):
        self.setup_everything()


    # 6.1. An incorrect parameter generates an error
    #
    # This is a utility function for carrying out a range of tests.  It resets
    # the configuration, sets the parameter named by 'param' to value, then
    # tries to start the replicator.  It expects to get an exception, whose
    # message should have the message id 'msgid'.

    def check_param(self, param, value, msgid):
        reset_configuration()
        config.__dict__[param] = value
        try:
            self.initialize_replicator()
        except:
            (_, msg, _) = sys.exc_info()
            if isinstance(msg, message.message):
                assert msg.id == msgid, \
                       ("Set parameter %s to '%s': expected message %d "
                        "but got %d (%s)"
                        % (param, value, msgid, msg.id, str(msg)))
            else:
                self.fail("Set parameter %s to '%s': expected message %d but "
                          "got '%s' instead." % (param, value, msgid, msg))
        else:
            self.fail("Set parameter %s to '%s': expected message %d but "
                      "didn't get it."
                      % (param, value, msgid))


    # 6.2. Basic errors in parameters are caught quickly
    #
    # Basic errors in parameters (wrong type, wrong format) should be caught
    # quickly.
    #
    # This is a table of (parameter name, bogus value, message id of expected
    # error).

    bogus_parameters = [
        ('dt_name', 'not a defect tracker', 1000),
        # Regression for job000170:
        ('administrator_address', 'invalid e-mail address', 202),
        ('changelist_url', -1, 207),
        ('changelist_url', "http://invalid/%d/%s", 210),
        ('changelist_url', "http://invalid/no/format/specifier", 210),
        ('changelist_url', "http://invalid/%d/%%/%%%", 210),
        ('log_file', -1, 208),
        ('log_level', 'not an int', 204),
        ('p4_client_executable', -1, 207),
        # Regression test for job000158:
        ('p4_client_executable', 'no such file', 705),
        ('p4_user', None, 207),
        ('p4_password', -1, 207),
        ('p4_port', None, 207),
        # Regression test for job000158, job000202:
        ('p4_port', '127.0.0.1:9999', 707),
        ('p4_server_description', -1, 207),
        ('poll_period', 'not an int', 204),
        ('replicator_address', 'invalid@e-mail@address', 202),
        ('rid', -1, 207),
        ('rid', '0abc', 209),
        ('rid', 'ab-c', 209),
        ('sid', -1, 207),
        ('sid', 'abcdefg+z', 209),
        ('smtp_server', -1, 207),
        ('start_date', '2001-02-03 24-00-00', 201),
        ('replicate_p', 'not a function', 203),
        ('closed_state', -1, 208),
        ('replicated_fields', 'not a list', 205),
        ('replicated_fields', ['not', 'a', 'list', 'of', 'strings', 0], 206),
        ]

    def test_parameters(self):
        import check_config
        for (param, value, msgid) in self.bogus_parameters:
            self.check_param(param, value, msgid)


    # 6.3. Basic errors in DT parameters are caught quickly
    #
    # As 6.2, but picks a set of tests based on config.dt_name.

    bogus_TeamTrack_parameters = [
        ('teamtrack_server', None, 207),
        ('teamtrack_user', None, 207),
        ('teamtrack_password', -1, 207),
        ('closed_state', 'not a TeamTrack state', 401),
        ('replicated_fields', ['PRIORITY', 'not a TeamTrack state'], 403),
        ('replicated_fields', ['OWNER'], 404),
        ('replicated_fields', ['TITLE'], 404),
        ('replicated_fields', ['STATE'], 404),
        ('replicated_fields', ['PRIORITY', 'PRIORITY'], 405),
        ('replicated_fields', ['MULTISELECT'], 406),
        # Regression for job000003, job000140:
        ('replicated_fields', ['CODE'], 407),
        ]

    bogus_Bugzilla_parameters = [
        ('bugzilla_directory', 'not a directory', 303),
        ('bugzilla_directory', '/', 304),
        ('closed_state', 'not a Bugzilla state', 301),
        ('dbms_database', -1, 207),
        ('dbms_host', -1, 207),
        # Regression for job000168:
        ('dbms_port', '1234', 204),
        ('dbms_user', -1, 207),
        ('dbms_password', -1, 207),
        # By using fake Perforce client executables we can check that
        # unsupported client and server versions are detected.
        ('p4_client_executable', './fake_p4.py', 704),
        ('p4_client_executable', './fake_p4d.py', 834),
        ('replicated_fields', ['bug_status'], 311),
        ('replicated_fields', ['assigned_to'], 311),
        ('replicated_fields', ['short_desc'], 311),
        ('replicated_fields', ['resolution'], 311),
        ('replicated_fields', ['longdesc', 'longdesc'], 312),
        ('replicated_fields', ['not a Bugzilla field'], 313),
        ]

    def test_dt_parameters(self):
        params = getattr(self, 'bogus_%s_parameters' % config.dt_name)
        for (param, value, msgid) in params:
            self.check_param(param, value, msgid)


    # 6.4. Parameter errors are caught by the defect tracker
    #
    # Like 6.2 and 6.3 this test sets a parameter to an incorrect value.  In
    # this case the error is caught by the defect tracker, so a message object
    # isn't returned, but rather a string, which we must test directly rather
    # than by message id.

    erroneous_TeamTrack_parameters = [
        ('teamtrack_server',
         'host.invalid',
         'TeamShare API error',
         'SOCKET_CONNECT_FAILED: Socket Connect failed.'),
        ('teamtrack_user',
         'invalid user',
         'TeamShare API error',
         'SERVER_ERROR: Authentication Failed.  Invalid user id or password'),
        ('teamtrack_password',
         'invalid password',
         'TeamShare API error',
         'SERVER_ERROR: Authentication Failed.  Invalid user id or password'),
        ]

    erroneous_Bugzilla_parameters = [
        ('dbms_host',
         'host.invalid',
         '_mysql.OperationalError',
         '(2005, "Unknown MySQL Server Host \'host.invalid\' (2)")'),
        ('dbms_database',
         'invalid',
         '_mysql.OperationalError',
         '(1049, "Unknown database \'invalid\'")'),
        # ('dbms_port',
	#  25,
        #  '_mysql.OperationalError',
	#  -1),
        ('dbms_password',
	 'not the Bugzilla password',
         '_mysql.OperationalError',
	 '(1045, "Access denied for user:'),
        ('dbms_user',
	 'not the Bugzilla user',
         '_mysql.OperationalError',
	 '(1045, "Access denied for user:'),
        ]

    def test_dt_errors(self):
        params = getattr(self, 'erroneous_%s_parameters' % config.dt_name)
        for (param, value, error, message_text) in params:
            reset_configuration()
            config.__dict__[param] = value
            try:
                self.initialize_replicator()
            except:
                (err, msg, _) = sys.exc_info()
                assert str(err) == error, \
                       ("Set parameter %s to '%s': expected error '%s' "
                        "but got error '%s'."
                        % (param, value, error, str(err)))
                assert str(msg)[0:len(message_text)] == message_text, \
                       ("Set parameter %s to '%s': expected error message "
                        "'%s' but got '%s'."
                        % (param, value, message_text, msg))
            else:
                self.fail("Set parameter %s to %s: expected error '%s' but "
                          "there was no error."
                          % (param, value, error))



# 7. TEST CASE: EXISTING JOB IN PERFORCE
#
# The replicator should refuse to start if there's a job in Perforce.
#
# This is a regression test for job000219 and job000240.

class existing_job(p4dti_base):
    def setUp(self):
        self.setup_everything()

    def runTest(self):
        j = self.p4.run('job -o')[0]
        j['Description'] = 'Test job'
        self.p4.run('job -i', j)
        self.check_startup_error("P4DTI Initialization error", 1001)


# 8. TEST CASE: RECENT START DATE
#
# When start_date is set to the current time, no issues should be replicated
# when the replicator starts.  Similarly, refreshing Perforce has no effect.
# But you can set start_date back in time and refresh Perforce, this time with
# effect.
#
# This is a regression test for job000047, job000050, job000221.

class recent_start_date(p4dti_base):
    def setUp(self):
        self.setup_everything()

    def runTest(self):
        config.start_date = time.strftime("%Y-%m-%d %H:%M:%S",
                                          time.gmtime(time.time()))
        self.initialize_replicator()

        # When we poll, nothing should happen.
        self.clear_log()
        self.r.poll()
        self.expected([])

        # Nor when we refresh.
        self.clear_log()
        self.r.refresh_perforce_jobs()
        self.expected([])

        # The databases should report consistent.
        self.check_consistency()

        # Now set start date back in time and try refreshing again.
        config.start_date = "1971-01-01 00:00:00"
        self.initialize_replicator()
        self.clear_log()
        self.r.refresh_perforce_jobs()
        self.expected([803, 804, 812])
        self.expected_only_in_range(800, 1000, [803, 804, 812])
        self.expected_not([846, 847, 806, 811])

        # The databases should still report consistent.
        self.check_consistency()


# 9. TEST CASE: REPLICATING BY PROJECT
#
# This is a regression test for job000107 and job000112.

class replicate_by_project(p4dti_variant):
    def setUp(self):
        self.setup_everything()

    def TeamTrack(self):
        config.replicated_fields = ['PROJECTID']
        config.replicate_p = (lambda self: self['PROJECTID'] in [3,4])
        self.check_startup()
        jobs = self.p4.run('jobs')
        for j in jobs:
            assert j['Project'] in ['Elaborate editor', 'Clever compiler'], \
                   ("Job %s has project %s; shouldn't be replicated."
                    % (j['Job'], j['Project']))
        self.check_consistency()

    def Bugzilla(self):
        config.replicated_fields = ['product']
        config.replicate_p = (lambda self: self.bug['product'] == 'product 1')
        self.check_startup()
        jobs = self.p4.run('jobs')
        for j in jobs:
            assert j['Project'] in ['Elaborate editor', 'Clever compiler'], \
                   ("Job %s has project %s; shouldn't be replicated."
                    % (j['Job'], j['Project']))
        self.check_consistency()


# 10. TEST CASE: ISSUE LIFE CYCLE

class issue_lifecycle(p4dti_variant):
    def setUp(self):
        self.setup_everything()
        self.check_startup()

    def Bugzilla(self):
        # Submit a new bug.
        fields = {
            'reporter': 'nb@ravenbrook.com',
            'product': 'product 1',
            'version': 'unspecified',
            'component': 'component 1.1',
            'rep_platform': 'All',
            'op_sys': 'All',
            'priority': 'P1',
            'severity': 'critical',
            'assigned_to': '',
            'cc': '',
            'bug_file_loc': '',
            'short_desc': 'Test bug',
            'comment': 'Test long description.',
            'submit': '    Commit    ',
            'form_name': 'enter_bug',
            'Bugzilla_password': 'p4dtitest',
            'Bugzilla_login': 'nb@ravenbrook.com',
            }
        h = httplib.HTTP('swan.ravenbrook.com')
        h.set_debuglevel(100)
        h.putrequest('POST', '/bugzilla/post_bug.cgi')
        content = string.join(map(lambda f: urllib.quote_plus(f[0]) + '=' + urllib.quote_plus(f[1]), fields.items()), '&') + '\r\n'
        h.putheader('Content-Length', str(len(content)))
        h.endheaders()
        h.send(content)
        h.getreply()
        for l in h.getfile().readlines():
            print l
        self.fail('unimplemented')

    def TeamTrack(self):
        # We need to be logged in to TeamTrack as a user other than the
        # replicator in order to make a change that isn't ignored.
        import teamtrack
        s = teamtrack.connect('rb','', config.teamtrack_server)

        # Submit a new case.
        case = s.new_record(teamtrack.table['CASES'])
        case['TITLE'] = 'Issue lifecycle test'
        case['DESCRIPTION'] = 'This is a test issue for the lifecycle test.'
        case['ISSUETYPE'] = 1 # Bug report
        case['PROJECTID'] = 3 # Editor
        case['SEVERITY'] = 45
        # Regression test for job000182:
        case['ESTTIMETOFIX'] = 7199   # 1:59:59
        case['ACTTIMETOFIX'] = 445556 # 123:45:56
        issueid = case.submit("rb")

        # It gets replicated to Perforce.  This is a regression test for
        # job000233.
        self.check_replication_dt_to_p4(first_time = 1)

        # Check the job in Perforce.
        case2 = s.query(teamtrack.table['CASES'],
                        "TS_ISSUETYPE=%d AND TS_ISSUEID='%05d'"
                        % (case['ISSUETYPE'], issueid))[0]
        job = self.p4.run('job -o %s' % case2['P4DTI_JOBNAME'])[0]
        for (field, expected) in [('State', '_new'),
                                  ('Title', 'Issue lifecycle test'),
                                  ('Est._Time_to_Fix', '1:59:59'),
                                  ('Actual_Time_to_Fix', '123:45:56')]:
            assert job[field] == expected, \
                   ("Expected new job %s to have %s '%s', but found '%s'."
                    % (job['Job'], field, expected, job[field]))

        # Now assign the issue in TeamTrack.
        case2.transition('rb', 3)

        # It gets replicated to Perforce.
        self.check_replication_dt_to_p4()

        # The state should now be assigned.
        job2 = self.p4.run('job -o %s' % case2['P4DTI_JOBNAME'])[0]
        assert job2['State'] == 'assigned', \
               ("Expected new job %(Job)s to have state 'assigned', but it "
                "has state %(State)s." % job2)

        # Close the job in Perforce.
        job2['State'] = 'closed'
        self.p4.run('job -i', job2)

        # It gets replicated back to TeamTrack.  TeamTrack will change the
        # owner field and the replicator will replicate that back.  This is a
        # regression test for job000053.
        self.clear_log()
        self.r.poll()
        self.expected([846, 805, 825, 826])
        self.expected_only_in_range(800, 1000, [846, 805, 825, 826])
        job3 = self.p4.run('job -o %s' % case2['P4DTI_JOBNAME'])[0]
        assert job3['Owner'] != job2['Owner'], \
               ("Closed job %(Job)s but owner is still %(Owner)s." % job3)

        # Regression test for job000182.
        case3 = s.read_record(teamtrack.table['CASES'], case['ID'])
        assert case3['ESTTIMETOFIX'] == 7199
        assert case3['ACTTIMETOFIX'] == 445556
        self.check_consistency()


# RUNNING THE TESTS

def tests():
    # These class define several test cases, all starting with 'test_'
    test_collections = [
        normal,
        bogus_config,
        ]
    suite = unittest.TestSuite(map(lambda c: unittest.makeSuite(c,'test_'),
                                   test_collections))

    # These class define a single test case, called 'runTest'.
    test_cases = [
        existing_job,
        recent_start_date,
        replicate_by_project,
        issue_lifecycle,
        ]
    for t in test_cases:
        suite.addTest(t())
    return suite

if __name__ == "__main__":
    unittest.main(defaultTest="tests")


# A. REFERENCES
#
# [PyUnit] "PyUnit - a unit testing framework for Python"; Steve Purcell;
# <http://pyunit.sourceforge.net/>.
#
#
# B. DOCUMENT HISTORY
#
# 2001-03-14 GDR Created.
#
# 2001-03-15 GDR Added list of regression tests, discussion of limitations,
# snoping on e-mail, p4dti_variant class, test cases for existing jobs, recent
# start date and replicating by project.
#
# 2001-03-17 GDR Added test_dt_errors test case for errors caught by the defect
# tracker or its database.
#
#
# C. COPYRIGHT AND LICENCE
#
# This file is copyright (c) 2001 Perforce Software, Inc.  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1.  Redistributions of source code must retain the above copyright notice,
#     this list of conditions and the following disclaimer.
#
# 2.  Redistributions in binary form must reproduce the above copyright notice,
#     this list of conditions and the following disclaimer in the documentation
#     and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS AND CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
#
# $Id: //info.ravenbrook.com/project/p4dti/version/1.1/test/test_p4dti.py#2 $
